"""
Stage devices.
"""

device_types = {
    "asi": "Applied Scientific Instrumentation",
    "galvo": "Analog/Digital Device",
    "mcl": "Mad City Labs",
    "pi": "Physik Instrumente",
    "sutter": "Sutter Instruments",
    "synthetic": "Virtual Device",
    "tl_kcube_inertial": "ThorLabs KCube Inertial Device",
}
