"""
Camera devices.
"""

device_types = {
    "hamamatsu_lightning": "Hamamatsu ORCA Lightning",
    "hamamatsu_fire": "Hamamatsu ORCA Fire",
    "hamamatsu_flash": "Hamamatsu Flash 4.0",
    "photometics": "Photometrics Iris 15B",
    "synthetic": "Virtual Device",
}
